/*
 * Genital
 * Bart Trzynadlowski
 * Public domain
 */


/*
 * genital.c: Main program, memory space, RAM, and ROM code
 */

#include <go32.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/farptr.h>
#include "68k.h"
#include "genital.h"
#include "gfx.h"
#include "input.h"
#include "misc.h"
#include "save.h"
#include "vdp.h"
#include "vga.h"
#include "z80.h"

#define LINE_CLOCK_68K_HBLANK   40
#define LINE_CLOCK_68K          488 - 40
#define LINE_CLOCK_68K_TOTAL    LINE_CLOCK_68K_HBLANK + LINE_CLOCK_68K
#define LINE_CLOCK_Z80          LINE_CLOCK_68K_TOTAL / 2

struct genesis_struct   genesis;
struct control_struct   control;

int                     fskip = 0, fskip_count = 0;


void ShowHelp()
{
        printf("Genital Version %s by Bart Trzynadlowski: A Free Sega Genesis"
               " Emulator\n", VERSION);
        printf("Usage:     genital <SMD or BIN ROM> [options]\n");
        printf("Options:   -?,-h        Show this help text\n");
        printf("           -info        Show information about the ROM\n");
        printf("           -dump        Dump RAM, VRAM, CRAM, VSRAM, and Z80 RAM\n");
        printf("           -gen68k      Use Genital's 68000 emulator\n");
        printf("           -star        Use the Starscream 68000 emulator\n");
        printf("           -raze        Use the RAZE Z80 emulator\n");
        printf("           -fskip #     Frames to skip before drawing and blitting\n");
        printf("           -vsync       Uses vsync to impose a speed limit\n");
        printf("           -mode13      320x200 Mode 0x13 resolution\n");
        printf("           -modex       320x240 ModeX resolution\n");
        printf("           -usa         Force an American Sega Genesis system\n");
        printf("           -europe      Force a European Sega MegaDrive system\n");
        printf("           -japan       Force a Japanese Sega MegaDrive system\n");
        printf("           -z80         Attempt to actually emulate the Z80\n");
        printf("           -hint        Allow horizontal interrupts\n");
        printf("           -menacer     Emulate the Menacer via the mouse\n");
        printf("           -mrbtn #     Right mouse button value (hexadecimal)\n");
        printf("           -nopri       No Scroll/Sprite priorities\n");
        printf("           -nosprlimit  Disable sprite limit and masking\n");
        exit(0);
}

int FindOption(char *option, int p, int h, int u, int argc, char **argv)
{
        static int      t[] = { 0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0, 0, 0, 0, 0 };
        int             i;
        char            *c;
      
        if (argc > 128)
                argc = 128;     /* maximum this function can handle is 128 */

        /*
         * if p = 1 and h = 0 will find option and return decimal value of
         * argv[i+1], if h = 1 will read it as hex.
         * if p = 0 then it will return index of option in argv[], 0 not found
         * if u = 1 will return index of first occurance of untouched option
         */

        if (u)  /* find first untouched element */
        {
                for (i = 1; i < argc; i++)
                {
                        if (!t[i])      /* 0 indicates untouched */
                                return i;
                }
                return 0;
        }

        if (p)  /* find option and return integer value following it */
        {
                for (i = 1; i < argc; i++)
                {                
                        if (strcmp(argv[i], option) == 0)       /* found */
                        {
                                if (i >= argc)                  /* bounds! */
                                        return 0;
                                t[i + 1] = t[i] = 1;            /* touched */
                                if (!h)
                                        return atoi(argv[i + 1]);
                                else
                                        return strtoul(argv[i + 1], &c, 16);
                        }
                }
                return 0;                                       /* no match */
        }
        else    /* find option and return position */        
        {
                for (i = 1; i < argc; i++)
                {                
                        if (strcmp(argv[i], option) == 0)
                        {
                                t[i] = 1;
                                return i;       /* found! return position */
                        }
                }                                                          
                return 0;
        }
        return 0;
}


/*
 * Misc. code
 */

void SwapBuffer(unsigned char *buffer, int length)
{
        int     i, j;

        /* swap bytes in each word */
        for (i = 0; i < length; i += 2)
        {
                j = buffer[i];
                buffer[i] = buffer[i + 1];
                buffer[i + 1] = j;
        }
}


/*
 * ROM code
 */

void ConvertSMDToBIN(unsigned char *bin_block, unsigned char *smd_block)
{
        int     i, o = 1, e = 0;

        /* convert 16KB of SMD to BIN */
        for (i = 0; i < 8192; i++)
        {
                bin_block[o] = smd_block[i];                
                bin_block[e] = smd_block[i+8192];
                o += 2;
                e += 2;
        }
}        

void ROMLoad(char *file)
{
        FILE            *fp;
        int             i, j = 0, num_blocks, size;
        unsigned char   smd_block[16384];

        if ((fp = fopen(file, "rb")) == NULL)
        {
                fprintf(stderr, "Genital: File could not be opened: %s\n",
                        file);
                exit(1);
        }
        if ((genesis.rom = (unsigned char *) calloc(4194304 + 4,
        sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 4MB RAM for " \
                        "ROM\n");
                exit(1);
        }
        
        /* get size of file, and number of blocks in case it is SMD */
        fseek(fp, 0, SEEK_END);
        fgetpos(fp, (void *) &size);
        num_blocks = (size - 512) / 16384;
        fseek(fp, 0, SEEK_SET);
       
        /* load the file */
        if (strstr(file, ".bin") != NULL)       /* BIN */
        {
                if (size > 4194304)     /* too big? */
                {
                        fprintf(stderr, "Genital: ROM file excedes 4MB: %s\n",
                                file);
                        exit(1);
                }
                else
                        fread(genesis.rom, 1, size, fp);
        }
        else if (strstr(file, ".smd") != NULL)  /* SMD */
        {
                if (size > 4194304 + 512)       /* too big? */
                {
                        fprintf(stderr, "Genital: ROM file excedes 4MB: %s\n",
                                file);
                        exit(1);
                }                
                fseek(fp, 512, SEEK_SET);       /* skip SMD header */
                for (i = 0; i < num_blocks; i++)
                {
                        fread(smd_block, 1, 16384, fp);
                        ConvertSMDToBIN(&genesis.rom[j], smd_block);
                        j += 16384;     /* next 16KB chunk in rom[] */
                }
        }
        else
        {
                fprintf(stderr, "Genital: Unrecognized ROM format: %s\n",
                        file);
                exit(1);
        }

        fclose(fp);
}        

void ROMUnload()
{
        free(genesis.rom);
}

void ShowROMInfo(char *file)
{
        FILE            *fp;
        unsigned char   header[16384];
        unsigned char   smd[16384];
        char            string[49];

        /* check to see if it a BIN or SMD file */
        if ((strstr(file, ".bin") != NULL) || (strstr(file, ".smd") != NULL))
        {
                /* open file and read the header */
                if ((fp = fopen(file, "rb")) == NULL)
                {
                        fprintf(stderr, "Genital: File could not be opened: " \
                                "%s\n", file);
                        exit(1);
                }

                if (strstr(file, ".bin") != NULL)       /* bin file */
                        fread(header, 1, 512, fp);
                else                                    /* smd file */
                {
                        /* get first block of SMD */
                        fread(smd, 1, 512, fp);         /* skip header */
                        fread(smd, 1, 16384, fp);
                        ConvertSMDToBIN(header, smd);   /* decode */
                }

                printf("Showing information from cartridge header...\n");
                printf("                    File: %s\n", file);
                string[16] = '\0';
                strncpy(string, &header[0x100], 16);
                printf("                  System: %s\n", string);
                strncpy(string, &header[0x110], 16);
                printf("               Copyright: %s\n", string);
                string[48] = '\0';
                strncpy(string, &header[0x120], 48);
                printf("    Game name (domestic): %s\n", string);
                strncpy(string, &header[0x150], 48);
                printf("    Game name (overseas): %s\n", string);
                if (header[0x180] == 'G' && header[0x181] == 'M')
                        printf("           Software type: game\n");
                else if (header[0x180] == 'A' && header[0x181] == 'l')
                        printf("           Software type: educational\n");
                else
                        printf("           Software type: %c%c\n",
                               header[0x180], header[0x181]);
                string[11] = '\0';
                strncpy(string, &header[0x183], 11);
                printf("Product code and version: %s\n", string);
                printf("                Checksum: %02X%02X\n", header[0x18e],
                       header[0x18f]);
                string[16] = '\0';
                strncpy(string, &header[0x190], 16);
                printf("             I/O support: %s\n", string);
                printf("       ROM start address: %02X%02X%02X%02X\n",
                       header[0x1a0], header[0x1a1], header[0x1a2],
                       header[0x1a3]);
                printf("         ROM end address: %02X%02X%02X%02X\n",
                       header[0x1a4], header[0x1a5], header[0x1a6],
                       header[0x1a7]);
                string[20] = '\0';
                strncpy(string, &header[0x1bc], 20);                
                printf("              Modem data: %s\n", string);
                string[40] = '\0';
                strncpy(string, &header[0x1c8], 40);
                printf("                    Memo: %s\n", string);
                string[3] = '\0';
                strncpy(string, &header[0x1f0], 3);
                printf("               Countries: %s\n", string);

                fclose(fp);
        }
        else
        {
                fprintf(stderr, "Genital: Unrecognized ROM format\n");
                exit(1);
        }
        exit(0);
}


/*
 * RAM code and memory (RAM, ROM) access handlers
 */

unsigned MemReadByte(unsigned addr)
{
        if (addr >= 0xe00000)  
                return genesis.ram[(addr ^ 1) & 0xffff];
        if (addr < 0x400000)
                return genesis.rom[addr ^ 1];
        return 0;
}

unsigned MemReadWord(unsigned addr)
{
        if (addr >= 0xe00000)
                return GetWord(genesis.ram + (addr & 0xfffe));
        else if (addr < 0x400000)
                return GetWord(genesis.rom + addr);
        return 0;
}

void MemWriteByte(unsigned addr, unsigned data)
{
        if (addr >= 0xe00000)   /* RAM */
                genesis.ram[(addr ^ 1) & 0xffff] = data;
}

void RAMInit()
{
        if ((genesis.ram = calloc(65536, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 64KB of RAM " \
                        "for work RAM\n");
                exit(1);
        }
}

void RAMDeinit()
{
        free(genesis.ram);
}


/*
 * System I/O Registers (including Z80 control ports and RAM)
 */

unsigned IOReadByte(unsigned addr)
{
        if (addr == 0xa10002 || addr == 0xa10003)   /* control pad 1 */
        {
                if (control.ctrl1_select == 1)
                        return (control.ctrl1 & 0xff);
                else
                        return (control.ctrl1 >> 8);
        }
        if (addr == 0xa10004 || addr == 0xa10005)   /* control pad 2 */
        {
                if (genesis.menacer)
                {
                        if (control.ctrl2_select == 1)
                                return (control.ctrl2 & 0xff);
                        else
                                return (control.ctrl1 >> 8);
                }
                else
                        return 0xff;
        }
        if (addr == 0xa10001)   /* version */
                return genesis.version;

        if (addr == 0xa11100)   /* Z80 BUSREQ */
        {
                if (genesis.busreq)     /* 0 means busreq (Z80 inactive) */
                        return 0;
                else
                        return 1;
        }

        /*
         * Z80 area
         */

        if ((addr >= 0xa00000) && (addr < 0xa08000))
        {
                if (genesis.busreq)     /* important: _only_ during busreq */
                        return Z80Read(addr & 0x7fff);
        }

        return 0;
}

unsigned IOReadWord(unsigned addr)
{
        if (addr == 0xa10002)   /* control pad 1 (word) */
        {
                control.ctrl1 &= 0xffff;
                if (control.ctrl1_select == 1)
                        return ((control.ctrl1 & 0xff) << 8) | (control.ctrl1 & 0xff);
                else
                        return (control.ctrl1 & 0xff00) | ((control.ctrl1 >> 8) & 0xff);
        }
        if (addr == 0xa10004)   /* control pad 2, Menacer (word) */
        {
                if (genesis.menacer)
                {
                        control.ctrl2 &= 0xffff;
                        if (control.ctrl2_select == 1)
                                return ((control.ctrl2 & 0xff) << 8) | (control.ctrl2 & 0xff);
                        else
                                return (control.ctrl2 & 0xff00) | ((control.ctrl2 >> 8) & 0xff);
                }
                else
                        return 0xffff;
        }

        if (addr == 0xa11100)   /* Z80 BUSREQ */
        {
                if (genesis.busreq)     /* 0 means busreq (Z80 inactive) */
                        return 0;
                else
                        return 0x100;
        }

        return 0;
}

void IOWriteByte(unsigned addr, unsigned data)
{        
        if (addr == 0xa10002 || addr == 0xa10003)   /* control pad 1 */
        {
                if (data & 0x40)        /* select high? */
                        control.ctrl1_select = 1;
                else
                        control.ctrl1_select = 0;
                return;
        }
        if (addr == 0xa10004 || addr == 0xa10005)   /* pad 2, Menacer */
        {
                if (genesis.menacer)
                {
                        if (data & 0x40)        /* select high? */
                                control.ctrl2_select = 1;
                        else
                                control.ctrl2_select = 0;                 
                }
                return;
        }


        /*
         * Z80 BUSREQ -- words are written to 0xa11100, I'm assuming bytes
         * would be written since we are modifying word bit 8.
         */

        if (addr == 0xa11100)
        {
                genesis.busreq = data & 1;      /* bit 8 (bit 0 of byte) */
                return;
        }

        /*
         * Z80 reset
         */

        if (addr == 0xa11200)
        {
                if (!(data & 1))        /* writing zero resets */
                        Z80Reset();
                return;
        }

        /*
         * Z80 RAM
         */

        if ((addr >= 0xa00000) && (addr < 0xa08000))
        {
                if (genesis.busreq)     /* important: _only_ during busreq */
                        Z80Write(addr & 0x7fff, data);
        }
}

void IOWriteWord(unsigned addr, unsigned data)
{
        if (addr == 0xa10002)   /* control pad 1 (word) */
        {
                if (data & 0x40 || data & 0x4000)       /* select high? */
                        control.ctrl1_select = 1;
                else
                        control.ctrl1_select = 0;
                return;
        }
        if (addr == 0xa10004)   /* control pad 2, Menacer (word) */
        {
                if (genesis.menacer)
                {
                        if (data & 0x40 || data & 0x4000)  /* select high? */
                                control.ctrl2_select = 1;
                        else
                                control.ctrl2_select = 0;
                }
                return;
        }

        /*
         * Z80 BUSREQ
         */

        if (addr == 0xa11100)
        {               
                if (data & 0x100)
                        genesis.busreq = 1;
                else
                        genesis.busreq = 0;
                return;
        }

        /*
         * Z80 reset
         */

        if (addr == 0xa11200)
        {
                if (!(data & 0x100))    /* writing zero to bit 8 resets */
                        Z80Reset();
                return;
        }
}

/*
 * Genesis emulator
 */

int main(int argc, char **argv)
{        
        FILE            *fp;
        int             i;
        unsigned        k;

        _farsetsel(_dos_ds);    /* FS holds far memory pointer */ 

        /* check command line for no options, process help */
        if (argc == 1)                                  /* show help */
                ShowHelp();
        if (FindOption("-?", 0, 0, 0, argc, argv) ||
            FindOption("-h", 0, 0, 0, argc, argv))
                ShowHelp();
        if (FindOption("-info", 0, 0, 0, argc, argv))   /* show ROM info */
        {
                if ((i = FindOption(NULL, 0, 0, 1, argc, argv)))
                        ShowROMInfo(argv[i]);
                else
                {
                        printf("Genital: Specify ROM path on command line\n");
                        exit(1);
                }
        }

        /* command line options and configure */
        fskip = fskip_count = FindOption("-fskip", 1, 0, 0, argc, argv);
        if (fskip < 0)  /* if below 0, make it at least 0 */
                fskip = fskip_count = 0;
                
        if (FindOption("-vsync", 0, 0, 0, argc, argv))  /* wait until vsync */
                genesis.vsync = 1;
        else
                genesis.vsync = 0;
        if (FindOption("-nopri", 0, 0, 0, argc, argv))  /* priorities? */
                genesis.pri = 0;
        else
                genesis.pri = 1;
        if (FindOption("-nosprlimit", 0, 0, 0, argc, argv)) 
                genesis.spr_limit = 0;                  /* sprite limit? */
        else
                genesis.spr_limit = 1;
        if (FindOption("-hint", 0, 0, 0, argc, argv))   /* h-ints? */
                genesis.hint = 1;
        else
                genesis.hint = 0;
        if (FindOption("-z80", 0, 0, 0, argc, argv))    /* z80? */
                genesis.z80_emu = 1;
        else
                genesis.z80_emu = 0;

        genesis.busreq = 1;     /* Genesis has bus at start-up */
        genesis.save_slot = 0;  

        /* set up the input devices */
        if (KeyboardInit())
        {
                KeyboardDeinit();
                printf("Genital: Failed to install keyboard driver\n");
                exit(1);
        }
        if (FindOption("-menacer", 0, 0, 0, argc, argv))        
        {
                if (MouseInit())
                {
                        KeyboardDeinit();
                        MouseDeinit();
                        printf("Genital: Microsoft Mouse Driver not "
                               "loaded\n");
                        exit(1);
                }
                genesis.menacer = 1;    /* yes, we have a Menacer */
        }
        else
                genesis.menacer = 0;
        if (!(genesis.mrbtn = FindOption("-mrbtn", 1, 1, 0, argc, argv)))
                genesis.mrbtn = 0x8;

        /* start... */
        if ((i = FindOption(NULL, 0, 0, 1, argc, argv)))
                ROMLoad(argv[i]);
        else
        {
               printf("Genital: Specify ROM path on command line\n");
               exit(1);
        }
        if (FindOption("-usa", 0, 0, 0, argc, argv))    /* which country */
                genesis.version = 0xa0;
        else if (FindOption("-europe", 0, 0, 0, argc, argv))
                genesis.version = 0xe0;
        else if (FindOption("-japan", 0, 0, 0, argc, argv))
                genesis.version = 0x20;
        else    /* autodetect */
        {
                for (i = 0; i < 3; i++) /* look for USA first */
                        if (genesis.rom[0x1f0 + i] == 'U')
                        {
                                genesis.version = 0xa1;
                                goto auto_done;
                        }
                for (i = 0; i < 3; i++) /* ... then Europe */
                        if (genesis.rom[0x1f0 + i] == 'E')
                        {
                                genesis.version = 0xe0;
                                goto auto_done;
                        }
                for (i = 0; i < 3; i++) /* ... Japan */
                        if (genesis.rom[0x1f0 + i] == 'J')
                        {
                                genesis.version = 0x20;
                                goto auto_done;
                        }
                genesis.version = 0xa1; /* default is USA */
                auto_done:
        }                                 
        RAMInit();
        /* Z80 is initialized before the 68K to provide a proper z80_ram */
        if (FindOption("-raze", 0, 0, 0, argc, argv))
                Z80Init(RAZE_Z80);
        else
                Z80Init(DEFAULT_Z80);
        if (FindOption("-gen68k", 0, 0, 0, argc, argv))
                CPUInit(GENITAL68K_CPU);
        else if (FindOption("-star", 0, 0, 0, argc, argv))
                CPUInit(STARSCREAM_CPU);
        else
                CPUInit(DEFAULT_CPU);    
        VDPInit();

        if (FindOption("-mode13", 0, 0, 0, argc, argv))
                GFXInit(MODE13);
        else if (FindOption("-modex", 0, 0, 0, argc, argv))
                GFXInit(MODEX);
        else
                GFXInit(MODE13);
        GFXPrint("GENITAL VERSION %s", VERSION);
        while (!control.quit)
        {
                VDPArmHIntTimer();
                for (i = 0; i < 224; i++)
                {                                                
                        VDPUpdateHVCounter(i);
                        VDPHBlankStart(i);
                        CPURun(LINE_CLOCK_68K_HBLANK);
                        VDPHBlankEnd();
                        CPURun(LINE_CLOCK_68K);
                        Z80Run(LINE_CLOCK_Z80);
                        if (fskip_count == 0)
                                DrawLine(i);
                }
                VDPVBlankStart();
                Z80Interrupt(0xff);
                for (i = i; i < 224 + 38; i++)
                {
                        VDPUpdateHVCounter(i);
                        CPURun(LINE_CLOCK_68K_TOTAL);
                        Z80Run(LINE_CLOCK_Z80);
                }
                if (fskip_count == 0)
                {
                        GFXDrawText();  /* update text, then screen */
                        BlitScreen();
                }
                VDPVBlankEnd();
                vdp.status_register ^= 0x10;

                /* update fskip counter */
                fskip_count--;
                if (fskip_count < 0)
                        fskip_count = fskip;    /* reload */

                k = KeyboardGetKey();           /* get key */
                if (k == 0xf)                   /* TAB: reset 68K */
                {
                        CPUReset();
                        GFXPrint("68000 RESET");
                }
                else if (k == 0x3a)             /* Caps Lock: reset Z80 */
                {
                        if (genesis.z80_emu)
                        {
                                Z80Reset();
                                GFXPrint("Z80 RESET");
                        }
                        else
                                GFXPrint("Z80 NOT ENABLED");
                }
                else if (k == 0x4a)             /* -: frameskip down */
                {
                        fskip--;
                        if (fskip < 0)          /* zero is low limit */
                                fskip = 0;
                        GFXPrint("FRAMESKIP: %d", fskip);
                }
                else if (k == 0x4e)             /* +: frameskip up */
                {
                        fskip++;
                        if (fskip > 20)         /* 20 is high limit */
                                fskip = 20;
                        GFXPrint("FRAMESKIP: %d", fskip);
                }
                else if (k == 0x3b)             /* F1: vsync toggle */
                {
                        if (genesis.vsync)
                        {
                                genesis.vsync = 0;
                                GFXPrint("VSYNC: OFF");
                        }
                        else
                        {
                                genesis.vsync = 1;
                                GFXPrint("VSYNC: ON");
                        }
                }
                else if (k == 0x3c)             /* F2: priority toggle */
                {
                        if (genesis.pri)
                        {
                                genesis.pri = 0;
                                GFXPrint("SCROLL/SPRITE PRIORITIES: OFF");
                        }
                        else
                        {
                                genesis.pri = 1;
                                GFXPrint("SCROLL/SPRITE PRIORITIES: ON");
                        }
                }
                else if (k == 0x3d)             /* F3: h-int toggle */
                {
                        if (genesis.hint)
                        {
                                genesis.hint = 0;
                                GFXPrint("HORIZONTAL INTERRUPTS: OFF");
                        }
                        else
                        {
                                genesis.hint = 1;
                                GFXPrint("HORIZONTAL INTERRUPTS: ON");
                        }
                }
                else if (k == 0x3e)             /* F4: sprite limit */
                {
                        if (genesis.spr_limit)
                        {
                                genesis.spr_limit = 0;
                                GFXPrint("SPRITE LIMIT: OFF");
                        }
                        else
                        {
                                genesis.spr_limit = 1;
                                GFXPrint("SPRITE LIMIT: ON");
                        }
                }
                else if (k == 0x3f)             /* F5: save state */
                        GSTSave(argv[FindOption(NULL, 0, 0, 1, argc, argv)]);
                else if (k == 0x40)             /* F6: toggle Z80 */
                {
                        if (genesis.z80_emu)
                        {
                                genesis.z80_emu = 0;
                                GFXPrint("Z80: OFF");
                        }
                        else
                        {
                                genesis.z80_emu = 1;
                                GFXPrint("Z80: ON");
                        }
                }                        
                else if (k == 0x41)             /* F7: load state */
                        GSTLoad(argv[FindOption(NULL, 0, 0, 1, argc, argv)]);        
                else if (k == 1)                /* escape */
                        control.quit = 1;
                else if ((k - 2) <= 9)          /* change save slot */
                {
                        int     slot[] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 0 };
                        genesis.save_slot = slot[k - 2];
                        GFXPrint("SAVE SLOT: %d", genesis.save_slot);
                }
                else if (k == 0x57)             /* F11: Mode 0x13 */
                {
                        GFXSwitch(MODE13);
                        GFXPrint("VIDEO MODE: 320x200 MODE 0x13 (LINEAR)");
                }
                else if (k == 0x58)             /* F12: ModeX */
                {
                        GFXSwitch(MODEX);
                        GFXPrint("VIDEO MODE: 320x240 MODEX (PLANAR)");
                }
                else if (k == 0x29)             /* `: debug */
                        CPUDebug();
        }
        GFXDeinit();

        /* dump memory */
        if (FindOption("-dump", 0, 0, 0, argc, argv))
        {
                if ((fp = fopen("ram.out", "wb")) != NULL)
                {
                        SwapBuffer(genesis.ram, 65536);
                        fwrite(genesis.ram, 1, 65536, fp);
                        fclose(fp);
                        printf("Genital: RAM dumped to RAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump RAM\n");
                if ((fp = fopen("vram.out", "wb")) != NULL)
                {
                        SwapBuffer(vdp.vram, 65536);
                        fwrite(vdp.vram, 1, 65536, fp);
                        fclose(fp);
                        printf("Genital: VRAM dumped to VRAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump VRAM\n");
                if ((fp = fopen("cram.out", "wb")) != NULL)
                {
                        SwapBuffer(vdp.cram, 128);
                        fwrite(vdp.cram, 1, 128, fp);
                        fclose(fp);
                        printf("Genital: CRAM dumped to CRAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump CRAM\n");
                if ((fp = fopen("vsram.out", "wb")) != NULL)
                {
                        SwapBuffer(vdp.vsram, 80);
                        fwrite(vdp.vsram, 1, 80, fp);
                        fclose(fp);
                        printf("Genital: VSRAM dumped to VSRAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump VSRAM\n");
                if ((fp = fopen("z80ram.out", "wb")) != NULL)
                {
                        fwrite(genesis.z80_ram, 1, 8192, fp);
                        fclose(fp);
                        printf("Genital: Z80 RAM dumped to Z80RAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump Z80 RAM\n");

        }

        /* quit */
        VDPDeinit();
        RAMDeinit();
        ROMUnload();
        KeyboardDeinit();
        if (FindOption("-menacer", 0, 0, 0, argc, argv))
                MouseDeinit();

        return 0;
}


