/*
 * gfx.c: Graphics rendering engine
 */

#include <pc.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/farptr.h>
#include "genital.h"
#include "gfx.h"
#include "misc.h"
#include "tile.h"
#include "vdp.h"
#include "vga.h"

/*
 * Screen is 320x224. The video buffer has each line as: 8 + 320 + 8. There
 * are also 8 lines above and below the virtual screen. These unused 8 pixel
 * tall or wide areas are for h- and v-scrolling.
 */

#define BUFFER_SIZE     ((8 + 320 + 8) * 224) + ((8 + 320 + 8) * (8 + 8))
#define BUFFER_CELL_WIDTH       ((8 + 320 + 8) >> 3)
#define BUFFER_WIDTH    (BUFFER_CELL_WIDTH << 3)
unsigned char           *video_buffer;  /* ~80640 bytes */
char                    gfxprint_text[256];
int                     gfxprint_timer = 0;
int                     gfx_res;

void             (*BlitScreen)(int);


void DrawSprites(int line, int pr)
{
        int             i, tile, x, y, yline, xlimit, height, tile_x, tile_y;
        unsigned char   *video, *spr;

        for (i = vdp.sprite_count - 1; i >= 0; --i)
        {                
                spr = &vdp.sprite_addr[vdp.sprite_order[i] << 3];
                tile = GetWord(spr + 4);

                /* don't draw if incorrect priority */
                if (((tile & 0x8000) >> 15) != pr)
                        goto loop_end;

                x = (GetWord(spr + 6) & 0x1ff) - 0x80;
                if (vdp.reg[12] & 6)    /* interlace 2 */
                        y = ((GetWord(spr + 0) & 0x3fe) >> 1) - 0x80;
                else
                        y = (GetWord(spr + 0) & 0x1ff) - 0x80;

                /* if (line - y) >= 0, then sprite is above or on this line
                   and may be visible. a negative value means the sprite is
                   not on the screen */
                if ((yline = line - y) < 0)
                        goto loop_end;

                xlimit = ((spr[2^1] << 1) & 0x18) + x;
                height = (spr[2^1] & 0x3);

                 /* do if on this line */
                if ((xlimit > -8) && (x < 320) && (yline >= 0) && (yline <= (height << 3) + 7))
                {                                         
                        tile_y = yline & 7;     /* line = tile line */

                        /* modify the actual tile entry to get the proper
                           vertical tile */                        
                        if (tile & 0x1000)      /* vflip */
                                tile += height - (yline >> 3);
                        else
                                tile += (yline >> 3);
                        height++;                                               

                        if (tile & 0x800)       /* hflip */
                        {                                                        
                                video = (line * ((BUFFER_CELL_WIDTH) << 3)) + video_buffer + xlimit;
                                for (tile_x = xlimit; tile_x >= x; tile_x -= 8)
                                {                                        
                                        if ((tile_x > -8) && (tile_x < 320))
                                                DrawSpriteTile(tile, tile_y, video);
                                        tile += height;
                                        video -= 8; 
                                }
                        }
                        else
                        {
                                video = (line * ((BUFFER_CELL_WIDTH) << 3)) + video_buffer + x;
                                for (tile_x = x; tile_x <= xlimit; tile_x += 8)
                                {
                                        if ((tile_x > -8) && (tile_x < 320))
                                                DrawSpriteTile(tile, tile_y, video);
                                        tile += height;
                                        video += 8;
                                }
                        }
                }
        loop_end:
        }
        
}

void DrawWindow(int line, int pr)
{        
        int             i;
        unsigned char   *ntab, *video, width = (vdp.reg[12] & 1) ? 64 : 32;

        /* only do if on line and visible */
        if (!vdp.reg[17] && !vdp.reg[18])
                return;
        if (!vdp.reg[17] && (vdp.reg[18] & 0x1f) >= 28)
                return; /* if out of v && h==0 then don't show */
        if (!(((line >> 3) >= vdp.window_y) && ((line >> 3) <= vdp.window_ylimit)))
                return;

        video = (line * ((BUFFER_CELL_WIDTH) << 3)) + video_buffer + (vdp.window_x << 3);
        ntab = vdp.window_addr + (((line >> 3) * width) << 1) + (vdp.window_x << 1);

        for (i = vdp.window_x; i <= vdp.window_xlimit; i++)
        {
                if (((GetWord(ntab) >> 15) & 1) == pr)
                        DrawTile((line & 7), ntab, video);
                ntab += 2;
                video += 8;
        }               
}

void DrawLine(int line)
{
        if (!(vdp.reg[1] & 0x40))       /* display is off, don't draw */
                return;

        VDPCalcSpriteOrder(line);

        if (genesis.pri)
        {
                #define PRI 1

                /* Scroll B low */
                #define SCROLL 1
                #define SOLID 1
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                /* Scroll A low */
                #define SCROLL 0
                #define SOLID 1
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                DrawWindow(line, 0);
                DrawSprites(line, 0);

                /* Scroll B high */
                #define SCROLL 1
                #define SOLID 0
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                /* Scroll A high */
                #define SCROLL 0
                #define SOLID 0
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL
                   
                DrawWindow(line, 1);
                DrawSprites(line, 1);

                #undef PRI
        }
        else
        {
                #define PRI 0

                /* Scroll B */
                #define SCROLL 1
                #define SOLID 1
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                /* Scroll A low */
                #define SCROLL 0
                #define SOLID 0
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                DrawWindow(line, 0);
                DrawWindow(line, 1);
                DrawSprites(line, 0);
                DrawSprites(line, 1);

                #undef PRI
        }
} 

void Mode13BlitScreen()
{
        int             i;
        unsigned int    pc_vram = 0xa0000;
        unsigned char   *video;

        video = video_buffer;

        if (genesis.vsync)      /* wait until vsync */
                while (!(inportb(INPUT_STATUS) & 8))
                        ;
                
        if (vdp.reg[12] & 0x1)  /* H40 */
        {
                for (i = 0; i < 200; i++)
                {
                        FastMemcpy((void *) video, (void *) pc_vram, 320);
                        pc_vram += 320;         /* to next line in PC VRAM */
                        video += 320 + 8 + 8;   /* ... in video buffer */
                }
        }
        else                    /* H32 */
        {
                for (i = 0; i < 200; i++)
                {
                        FastMemset((void *) pc_vram, 255, 32);
                        pc_vram += 32;          /* center screen */
                        FastMemcpy((void *) video, (void *) pc_vram, 256);
                        pc_vram += 256;
                        FastMemset((void *) pc_vram, 255, 32);
                        pc_vram += 32;          /* next line... */
                        video += 320 + 8 + 8;
                }

        }
}

void ModeXBlitScreen()
{
        int             i, j, k;
        unsigned        pc_vram = 0xa0000;
        unsigned char   *video;

        video = video_buffer;

        if (genesis.vsync)      /* wait until vsync */
                while (!(inportb(INPUT_STATUS) & 8))
                        ;
                
        /*
         * ModeX:
         *
         * This blitting routine is slow and ugly, but this is my first time
         * working with ModeX and I can't think of a better way.
         */

        if (vdp.reg[12] & 0x1)  /* H40 */
        {
                for (i = 0; i < 224; i++)
                {
                        for (j = 0; j < 4; j++)
                        {
                                /* plane */
                                outportb(0x3c4, 0x02);
                                outportb(0x3c5, 0x01 << j);
                                /* start at 0, every 4 bytes because we're
                                   doing one plane at a time */
                                for (k = j; k < 320; k += 4)
                                {
                                        _farnspokeb(pc_vram, video[k]);
                                        pc_vram++;
                                }
                                pc_vram -= 320 / 4; /* reset for next plane */
                        }
                        pc_vram += 320 / 4;
                        video += 320 + 8 + 8;   /* next line */
                }
        }
        else                    /* H32 */
        {
                for (i = 0; i < 224; i++)
                {
                        for (j = 0; j < 4; j++)
                        {
                                /* plane */
                                outportb(0x3c4, 0x02);
                                outportb(0x3c5, 0x01 << j);
                                /* start at 0, every 4 bytes because we're
                                   doing one plane at a time */
                                for (k = 0; k < 32; k += 4)
                                {
                                        _farnspokeb(pc_vram, 255);
                                        pc_vram++;
                                }
                                for (k = j; k < 256; k += 4)
                                {
                                        _farnspokeb(pc_vram, video[k]);
                                        pc_vram++;
                                }
                                for (k = 0; k < 32; k += 4)
                                {
                                        _farnspokeb(pc_vram, 255);
                                        pc_vram++;
                                }
                                pc_vram -= 320 / 4; /* reset for next plane */
                        }
                        pc_vram += 320 / 4;
                        video += 320 + 8 + 8;   /* next line */
                }
        }
}

void GFXPrint(char *input, ...)
{
        va_list arg;
        va_start(arg, input);
        vsprintf(gfxprint_text, input, arg);
        va_end(arg);
        gfxprint_timer = 70;   /* display message for 70fps */
}

void GFXDrawText()
{
        int i = 0, x, y, c = 254, j, k, offset;
        unsigned character;   
        unsigned char mask, bits;
        #include "font.h"

        if (!gfxprint_timer)
                return;

        gfxprint_timer--;

        x = 5;
        if (gfx_res == MODEX)
                y = 211;
        else
                y = 187;                

        offset = (y * (BUFFER_WIDTH)) + x;         /* calculate start offset */
        while (gfxprint_text[i] != '\0' && i < (BUFFER_CELL_WIDTH - 1))
        {
                /* use the character to find its font from BIOS font table */
                character = /*0xffa6e*/ gfxprint_text[i] * 8;

                for (j = 0; j < 8; j++)
                {
                        mask = 0x80;
                        bits = font[character];
                
                        for (k = 0; k < 8; k++)
                        {
                                if (bits & mask)
                                        video_buffer[offset] = c;
                                offset++;
                                mask = mask >> 1;
                        }
                        /* return to x, next y pos */
                        offset = offset - 8 + BUFFER_WIDTH;
                        character++;
                }
                offset -= (8 * BUFFER_WIDTH) - 8;
                i++;    
        }
}


void GFXSwitch(int res)
{
        gfx_res = res;
        if (res == MODE13)
                BlitScreen = &Mode13BlitScreen;
        else if (res == MODEX)
                BlitScreen = &ModeXBlitScreen;

        GraphicsSetMode(res);
}

void GFXInit(int res)
{
        if ((video_buffer = calloc(BUFFER_SIZE, sizeof(unsigned char)))
            == NULL)
        {
                fprintf(stderr, "Genital: GFXInit() failed\n");
                exit(1);
        }

        /* forward 8 lines and 8 pixels to the start of the displayed
           portion of the buffer */
        video_buffer += ((8 + 320 + 8) * 8);// + 8;

        if (res == MODE13)
                BlitScreen = &Mode13BlitScreen;
        else if (res == MODEX)
                BlitScreen = &ModeXBlitScreen;

        GraphicsInit(res);
}

void GFXDeinit()
{
        video_buffer -= ((8 + 320 + 8) * 8);// + 8;
        free(video_buffer);

        GraphicsDeinit();
}




